#
# Utilities
#

enquote <- function(x) as.call(list(as.name("quote"), x))

docall <- function(fun, args) {
    if ((is.character(fun) && length(fun) == 1) || is.name(fun))
        fun <- get(as.character(fun), env = .GlobalEnv, mode = "function")
    do.call("fun", lapply(args, enquote))
}


#
# Slave Loop Function
#

slaveLoop <- function(master) {
    repeat {
        msg <- recvData(master)
	cat(paste("Type:", msg$type, "\n"))

        if (msg$type == "DONE") {
            closeNode(master)
            break;
        }
        else if (msg$type == "EXEC") {
            value <- try(docall(msg$data$fun, msg$data$args))
            #**** use exception= if failure?
            value <- list(type = "VALUE", value = value, tag = msg$data$tag)
            sendData(master, value)
        }
    }
}


#
# Higher-Level Node Functions
#

closeNode <- function(node) UseMethod("closeNode")
closeNode.default <- function(node) {}

sendData <- function(node, data) UseMethod("sendData")
recvData <- function(node) UseMethod("recvData")

postNode <- function(con, type, value = NULL, tag = NULL) {
    sendData(con, list(type = type, data = value, tag = NULL))
}

stopNode <- function(n) {
    postNode(n, "DONE")
    closeNode(n)
}


#
#  Cluster Creation and Destruction
#

defaultClusterOptions <- NULL

#**** check valid cluster option

setDefaultClusterOptions <-
    function(port = 600011,
             master =  Sys.info()["nodename"],
             outfile = "/dev/null",
             rhome = Sys.getenv("R_HOME"),
             user = Sys.info()["user"],
             rshcmd = "ssh",
             rlibs = Sys.getenv("R_LIBS"),
             scriptdir = .path.package("snow"),
             rprog = file.path(rhome, "bin", "R")) {
    defaultClusterOptions <<- new.env(parent=NULL, hash=TRUE)
    for (n in ls()) assign(n, get(n), env = defaultClusterOptions)
}

addClusterOptions <- function(options, ...) {
    new <- list(...)
    if (! is.null(new)) {
        options <- new.env(parent = options)
        names <- names(new)
        for (i in seq(along = new))
            assign(names[i], new[[i]], env = options)
    }
    options
}

getClusterOption <- function(name, options = defaultClusterOptions)
    get(name, env = options)

stopCluster <- function(cl) UseMethod("stopCluster")

stopCluster.default <- function(cl)
    for (n in cl) stopNode(n)


#
# Cluster Functions
#

sendCall <- function(con, fun, args, return = TRUE)
    #**** mark node as in-call
    postNode(con, "EXEC", list(fun = fun, args = args, return = return))

recvResult <- function(con) recvData(con)$value

clusterCall  <- function(cl, fun, ...) {
    for (i in seq(along = cl))
        sendCall(cl[[i]], fun, list(...))
    lapply(cl, recvResult)
}

clusterApply <- function(cl, x, fun, ...) {
    if (length(cl) < length(x))
        stop("data lengh must be at most cluster size")
    for (i in seq(along = x))
        sendCall(cl[[i]], fun, c(list(x[[i]]), list(...)))
    lapply(cl[seq(along=x)], recvResult)
}


clusterEvalQ<-function(cl, expr)
    clusterCall(cl, eval, substitute(expr), env=.GlobalEnv)

clusterExport <- function(cl, list) {
    # do this with only one clusterCall--loop on slaves
    for (name in list) {
        clusterCall(cl, function(n, v) assign(n, v, env = .GlobalEnv),
                        name, get(name, env = .GlobalEnv))
    }
}

#
# Cluster SPRNG Support 
#
# adapted from rpvm (Li & Rossini)

clusterSetupSPRNG <- function (cl, seed = round(2^32 * runif(1)),
                            prngkind = "default", para = 0) 
{
    if (!is.character(prngkind) || length(prngkind) > 1)
        stop("'rngkind' must be a character string of length 1.")
    if (!is.na(pmatch(prngkind, "default")))
        prngkind <- "LFG"
    prngnames <- c("LFG", "LCG", "LCG64", "CMRG", "MLFG", "PMLCG")
    kind <- pmatch(prngkind, prngnames) - 1
    if (is.na(kind))
        stop(paste("'", prngkind, "' is not a valid choice", sep = ""))
    nc <- length(cl)
    invisible(clusterApply(cl, 0:(nc-1), initSprngNode, nc, seed, kind, para))
}

initSprngNode <- function (streamno, nstream, seed, kind, para) 
{
    if (! require(rsprng))
        stop("the `rsprng' package is needed for SPRNG support.")
    .Call("r_init_sprng", as.integer(kind), as.integer(streamno), 
        as.integer(nstream), as.integer(seed), as.integer(para))
    RNGkind("user")
}


#
# Parallel Functions
#

splitIndices <- function(nx, ncl) {
    batchsize <- if (nx %% ncl == 0) nx %/% ncl else 1 + nx %/% ncl
    batches <- (nx + batchsize - 1) %/% batchsize
    split(1:nx, rep(1:batches, each = batchsize)[1:nx])
}

splitIndices <- function(nx, ncl) {
    i <- 1:nx;
    structure(split(i, cut(i, ncl)), names=NULL)
}

clusterSplit <- function(cl, seq)
    lapply(splitIndices(length(seq), length(cl)), function(i) seq[i])

splitList <- function(x, ncl)
    lapply(splitIndices(length(x), ncl), function(i) x[i])

splitRows <- function(x, ncl)
    lapply(splitIndices(nrow(x), ncl), function(i) x[i,, drop=F])

splitCols <- function(x, ncl)
    lapply(splitIndices(ncol(x), ncl), function(i) x[,i, drop=F])

parLapply <- function(cl, x, fun, ...)
    docall(c, clusterApply(cl, splitList(x, length(cl)), lapply, fun, ...))

parRapply <- function(cl, x, fun, ...)
    docall(c, clusterApply(cl, splitRows(x,length(cl)), apply, 1, fun, ...))

parCapply <- function(cl, x, fun, ...)
    docall(c, clusterApply(cl, splitCols(x,length(cl)), apply, 2, fun, ...))

parMM <- function(cl, A, B)
    docall(rbind,clusterApply(cl, splitRows(A, length(cl)), get("%*%"), B))

parSapply <- function (cl, X, FUN, ..., simplify = TRUE, USE.NAMES = TRUE) 
{
    FUN <- match.fun(FUN)
    answer <- parLapply(cl,as.list(X), FUN, ...)
    if (USE.NAMES && is.character(X) && is.null(names(answer))) 
        names(answer) <- X
    if (simplify && length(answer) != 0) {
        common.len <- unique(unlist(lapply(answer, length)))
        if (common.len == 1) 
            unlist(answer, recursive = FALSE)
        else if (common.len > 1) 
            array(unlist(answer, recursive = FALSE),
                  dim = c(common.len, length(X)),
                  dimnames = list(names(answer[[1]]), names(answer)))
        else answer
    }
    else answer
}


#
#  Library Initialization
#

.First.lib <- function(libname, pkgname) {
    if (is.null(defaultClusterOptions))
	setDefaultClusterOptions()
}
